import torch
import sys
import traceback
from typing import Union, List


def get_scheduler(optimizer, n_steps, end_lr_factor=0.1, n_warmup_steps=None):
    if n_warmup_steps is None:
        n_warmup_steps = 0.05 * n_steps

    def lr_lambda(step):
        if step < n_warmup_steps:
            return step / n_warmup_steps
        else:
            return 1 - (1 - end_lr_factor) * min(
                (step - n_warmup_steps), n_steps - n_warmup_steps
            ) / (n_steps - n_warmup_steps + 1e-1)

    return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lr_lambda)


def get_wsd_scheduler(
    optimizer, n_steps, end_lr_factor=0.1, n_warmup_steps=None, percent_cooldown=0.1
):
    if n_warmup_steps is None:
        n_warmup_steps = 0.05 * n_steps

    def lr_lambda(step):
        if step < n_warmup_steps:
            return step / n_warmup_steps
        elif step < (1 - percent_cooldown) * n_steps:
            return 1
        else:
            return 1 - (1 - end_lr_factor) * min(
                (step - (1 - percent_cooldown) * n_steps),
                (1 - percent_cooldown) * n_steps,
            ) / (percent_cooldown * n_steps + 1e-2)

    return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lr_lambda)


def get_restarted_wsd_scheduler(
    optimizer,
    cur_step,
    n_steps,
    n_restart_steps,
    end_lr_factor=0.1,
    n_warmup_steps=None,
    percent_cooldown=0.1,
):
    if n_warmup_steps is None:
        n_warmup_steps = 0.05 * n_steps

    def lr_lambda(step):
        if step < n_restart_steps:
            scaling_factor = 0.9 * (step / n_restart_steps) + 0.1
        else:
            scaling_factor = 1.0

        true_step = step + cur_step

        if true_step < n_warmup_steps:
            return scaling_factor * (true_step / n_warmup_steps)
        elif true_step < (1 - percent_cooldown) * n_steps:
            return scaling_factor
        else:
            return scaling_factor * (
                1
                - (1 - end_lr_factor)
                * min(
                    (true_step - (1 - percent_cooldown) * n_steps),
                    (1 - percent_cooldown) * n_steps,
                )
                / (percent_cooldown * n_steps + 1e-2)
            )

    return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lr_lambda)


def is_iterable(x):
    try:
        _ = iter(x)
        return True
    except:
        return False


def opt(x):
    # print elements of dir(x) that don't start with _
    items = [item for item in dir(x) if not item.startswith("_")]
    for item in items:
        print(item)


def reload_module(module_name: Union[str, List[str]]):
    if isinstance(module_name, list):
        for m in module_name:
            reload_module(m)
        return
    else:
        assert isinstance(module_name, str)
    for x in list(sys.modules):
        if x.split(".")[0] == module_name:
            del sys.modules[x]
    module = __import__(module_name, fromlist=[""])
    return module


def get_str_for_nested_tensor(t):
    if isinstance(t, torch.Tensor):
        return f"Tensor{tuple(t.shape)}"
    elif isinstance(t, torch.nn.Parameter):
        return f"Parameter{tuple(t.shape)}"
    elif isinstance(t, np.ndarray):
        return f"np_array{tuple(t.shape)}"
    elif isinstance(t, list) or isinstance(t, tuple):
        return f"[{', '.join([get_str_for_nested_tensor(x) for x in t])}]"
    elif isinstance(t, dict):
        dict_str = "{"
        for k, v in t.items():
            dict_str += f"{k}: {get_str_for_nested_tensor(v)}, "
        dict_str = dict_str[:-2] + "}"
        return dict_str
    else:
        return str(t)


def see(t):
    # renders array shape and name of argument
    stack = traceback.extract_stack()
    filename, lineno, function_name, code = stack[-2]
    code = code.replace("see(", "")[:-1]
    print(">> " + code, end="")
    if isinstance(t, torch.Tensor):
        if t.dtype == torch.bool:
            t = t.half()
        if t.dtype == torch.long or t.dtype == torch.int:
            abs_t = t.abs()
            absmax = abs_t.max()
            absmin = abs_t.min()
            print(
                f": {str(tuple(t.shape))} | absmax: {absmax} | absmin: {absmin} | {t.device}, {t.dtype}"
            )
        else:
            avg = t.mean().item()
            std = t.std().item()
            absmax = t.abs().max().item()
            print(
                f": {str(tuple(t.shape))} | avg={avg:.2G} std={std:.2G} absmax={absmax:.2G} | {t.device}, {t.dtype}"
            )
    elif isinstance(t, np.ndarray):
        if t.dtype == bool:
            t = t.astype(np.float32)
        if np.issubdtype(t.dtype, np.integer):
            mode = np.bincount(t.flatten()).argmax()
            print(f": {str(tuple(t.shape))} | mode: {mode} | {t.dtype}")
        else:
            avg = t.mean()
            std = t.std()
            print(f": {str(tuple(t.shape))} | avg={avg:.2G} std={std:.2G} | {t.dtype}")
    elif (isinstance(t, list) or isinstance(t, tuple)) or isinstance(t, dict):
        print(": " + get_str_for_nested_tensor(t))
    else:
        print(": " + str(t))


def asee(t: torch.Tensor):
    # short for array-see, renders array as well as shape
    stack = traceback.extract_stack()
    filename, lineno, function_name, code = stack[-2]
    code = code.replace("asee(", "")[:-1]
    print("> " + code, end="")
    if isinstance(t, torch.Tensor):
        print(": " + str(tuple(t.shape)))
        # print('line: '+str(lineno))
        print("arr: " + str(t.detach().cpu().numpy()))
    elif isinstance(t, np.ndarray):
        print(": " + str(t.shape))
        # print('line: '+str(lineno))
        print("arr: " + str(t))
    else:
        print(t)
        # print('line: '+str(lineno))
    print()


def batched_bincount(x, dim, max_value):
    # From Guillaume Leclerc: https://discuss.pytorch.org/t/batched-bincount/72819/3
    target = torch.zeros(x.shape[0], max_value, dtype=x.dtype, device=x.device)
    values = torch.ones_like(x)
    target.scatter_add_(dim, x, values)
    return target


class TensorHistogramObserver:
    def __init__(self, min, max, bin_width, tensor_shape):
        """
        Useful for storing bin counts for multiple histograms in parallel
        E.G. one for each (layer, neuron) pair, in which case tensor_shape=(n_layers, n_neurons)
        and self.update would be called with tensors of shape (n_layers, n_neurons, n_samples).
        """

        self.min = min
        self.max = max
        self.bin_width = bin_width
        self.tensor_shape = tensor_shape

        self.boundaries = torch.arange(self.min, self.max + bin_width, bin_width)
        self.counts = torch.zeros(*tensor_shape, len(self.boundaries) - 1).int()

    def update(self, obs):
        assert obs.shape[:-1] == self.tensor_shape
        obs = obs.detach().cpu()

        # flatten all but last dimension
        obs_view = obs.view(-1, obs.shape[-1])
        # bucket ids with shape (product(obs.shape[:-1]), obs.shape[-1])
        flattened_bucket_ids = (
            torch.bucketize(
                obs_view.clamp(self.min, self.max - self.bin_width * 1e-4),
                self.boundaries,
                right=True,
            )
            - 1
        )
        self.counts += batched_bincount(
            flattened_bucket_ids, dim=-1, max_value=len(self.boundaries) - 1
        ).view(*obs.shape[:-1], -1)



from PIL import Image
import torch
import numpy as np
from scipy import stats

import math
import einops
import plotly.graph_objects as go
import plotly.express as px

import matplotlib.pyplot as plt


def is_tensor(x):
    return isinstance(x, torch.Tensor) or isinstance(x, torch.nn.parameter.Parameter)


def tensor_to_numpy(x):
    # if x is a tensor, convert to numpy array
    if isinstance(x, torch.Tensor):
        if x.device != torch.device("cpu"):
            x = x.cpu()
        if x.dtype == torch.bool:
            x = x.int()
        if x.dtype == torch.float16:
            x = x.float()
        if x.dtype == torch.float64:
            x = x.float()
        x = x.detach().cpu().numpy()
    if isinstance(x, torch.nn.parameter.Parameter):
        x = tensor_to_numpy(x.data)
    return x


def render_array(arr, scale: int = 1, raw_array=False):
    arr = tensor_to_numpy(arr)
    assert isinstance(arr, np.ndarray)
    arr = np.squeeze(arr)
    assert len(arr.shape) == 2
    if scale != 1:
        arr = np.kron(arr, np.ones((scale, scale)))

    if np.issubdtype(arr.dtype, np.floating):
        if arr.min() < 0.0 or arr.max() > 1.0:
            if arr.min() > -1e-2 and arr.max() < 1.0 + 1e-2:
                arr = np.clip(arr, 0, 1)
            else:
                arr = (arr - arr.min()) / (arr.max() - arr.min() + 1e-2)
        arr = (255 * arr // 1).astype(np.uint8)
    else:
        assert np.issubdtype(arr.dtype, np.integer)

    if raw_array is True:
        return arr
    else:
        return Image.fromarray(arr)


def render_array_w_sign(arr, scale: int = 1):
    pos = np.clip(arr, 0, None)
    neg = np.clip(arr, None, 0).abs()
    pos = render_array(pos / arr.abs().max(), scale=scale, raw_array=True)
    neg = render_array(neg / arr.abs().max(), scale=scale, raw_array=True)

    both = np.zeros(pos.shape + (3,), dtype=np.uint8)
    both[..., 0] = neg
    both[..., 1] = pos

    return Image.fromarray(both)


def str_arr_add(*args):
    """
    Casts tensors/nparrays to numpy string arrays and adds all items together,
    casting to string and broadcasting if necessary
    """
    if len(args) == 0:
        return ""
    args = list(args)
    for i, item in enumerate(args):
        item = tensor_to_numpy(item)
        if isinstance(item, np.ndarray):
            args[i] = item.astype(str)
    res = args[0]
    for item in args[1:]:
        res = np.core.defchararray.add(res, item)
    return res


def heatmap(
    arr,
    perm_0=None,
    perm_1=None,
    dim_names=("row", "col"),
    info_0=None,
    info_1=None,
    include_idx=(True, True),
    title=None,
    mask_0=None,
    mask_1=None,
    sort_0=None,
    sort_1=None,
):
    """
    arr: 2d numpy or torch array to render
    dim_names : (str, str), names of dim 0 and dim 1 respectively
    info_0, info_1 : dictionary of string keys to list of strings describing the indices of dim 0 and dim 1 respectively

    sort_0, sort_1: 1d arrays of indices to sort the rows and columns of the heatmap by
    """

    assert not (
        perm_0 is not None and sort_0 is not None
    ), "Cannot provide both perm_0 and sort_0"
    assert not (
        perm_1 is not None and sort_1 is not None
    ), "Cannot provide both perm_1 and sort_1"

    # convert arr to numpy array
    arr = tensor_to_numpy(arr)
    assert isinstance(arr, np.ndarray)

    # Create default title if none is provided
    if title is None:
        if dim_names == ("row", "col"):
            title = f"{arr.shape}"
        else:
            title = f"({dim_names[0]}, {dim_names[1]})"

    # get permutations from sort arrays
    if sort_0 is not None:
        assert len(sort_0.shape) == 1
        perm_0 = torch.tensor(tensor_to_numpy(sort_0)).topk(k=len(sort_0)).indices
    if sort_1 is not None:
        assert len(sort_1.shape) == 1
        perm_1 = torch.tensor(tensor_to_numpy(sort_1)).topk(k=len(sort_1)).indices

    # if permutations are not provided, use the identity permutation
    perm_0 = np.arange(arr.shape[0]) if perm_0 is None else tensor_to_numpy(perm_0)
    perm_1 = np.arange(arr.shape[1]) if perm_1 is None else tensor_to_numpy(perm_1)

    def construct_dim_info(
        dim_info: dict, dim_name: str, dim_len, perm, mask=None, include_idx=False
    ):
        dim_info = {} if dim_info is None else dim_info

        if include_idx is True:
            dim_info[f"{dim_name}"] = np.arange(dim_len)

        for k, v in dim_info.items():
            if is_tensor(v):
                dim_info[k] = tensor_to_numpy(v)
            else:
                if not isinstance(v, np.ndarray):
                    dim_info[k] = np.array(v)

        dim_info = {k: v[perm] for k, v in dim_info.items()}

        if mask is not None:
            mask = tensor_to_numpy(mask)[perm]
            for k, v in dim_info.items():
                dim_info[k] = v[mask]

        dim_info = str_arr_add(
            *[str_arr_add(k + ": ", v, "<br>") for k, v in dim_info.items()]
        )
        dim_info = np.array(dim_info).astype(str).tolist()
        return dim_info

    # Construct hovertemplate and dim info for each dimension (0 and 1)
    hovertemplate = ""
    if info_0 is not None or include_idx[0]:
        hovertemplate += "%{y}"
        info_0 = construct_dim_info(
            info_0, dim_names[0], arr.shape[0], perm_0, mask_0, include_idx[0]
        )
    if info_1 is not None or include_idx[1]:
        hovertemplate += "%{x}"
        info_1 = construct_dim_info(
            info_1, dim_names[1], arr.shape[1], perm_1, mask_1, include_idx[1]
        )
    hovertemplate += "val: %{z:.2f}<extra></extra>"

    # apply masks and permutations
    arr = arr[perm_0][:, perm_1]
    if mask_0 is not None:
        arr = arr[mask_0[perm_0]]
    if mask_1 is not None:
        arr = arr[:, mask_1[perm_1]]

    # Create the plotly.graph_objects figure
    layout = go.Layout(yaxis=dict(autorange="reversed"))

    fig = go.Figure(
        data=go.Heatmap(
            z=arr,
            y=info_0,
            x=info_1,
            hovertemplate=hovertemplate,
            colorscale="Viridis",
        ),
        layout=layout,
    )

    fig.update_layout(
        xaxis_title=f"{dim_names[1]} ({arr.shape[1]})",
        yaxis_title=f"{dim_names[0]} ({arr.shape[0]})",
        title=title,
    )
    fig.update_xaxes(showticklabels=False)
    fig.update_yaxes(showticklabels=False)

    return fig


def qq_plot(x, dist="norm", sparams=(), hovertext=None):
    x = x.squeeze()
    assert len(x.shape) == 1
    perm = x.topk(x.shape[0], largest=False).indices
    hovertext = np.array(hovertext)[perm] if hovertext is not None else None
    qq = stats.probplot(x[perm], dist=dist, sparams=sparams)
    x = np.array([qq[0][0][0], qq[0][0][-1]])
    fig = go.Figure()
    fig.add_scatter(x=qq[0][0], y=qq[0][1], mode="markers", hovertext=hovertext)
    fig.update_xaxes(title="theoretical quantiles")
    fig.update_yaxes(title="actual quantiles")

    fig.add_scatter(x=x, y=qq[1][1] + qq[1][0] * x, mode="lines")
    fig.layout.update(showlegend=False)
    fig.show()


def get_image_grid(images, width: int = -1, scale: int = 1):
    """images is a list of PIL.Image images"""

    assert scale >= 1

    def find_closest_factors(n):
        for i in range(int(np.sqrt(n).item()) + 1, 0, -1):
            if n % i == 0:
                return i, n // i

    image_array = np.stack([np.array(image) for image in images], axis=0)
    if width == -1:
        rows, columns = find_closest_factors(len(images))
    else:
        assert isinstance(width, int)
        rows = math.ceil(len(images) / width)
        columns = width
        empty_images_needed = rows * columns - len(images)

        # add the empty images
        image_array = np.concatenate(
            [
                image_array,
                np.zeros((empty_images_needed,) + image_array.shape[1:]).astype(
                    np.uint8
                ),
            ],
            axis=0,
        )

    image_arr = einops.rearrange(
        image_array, "(b1 b2) h w c -> (b1 h) (b2 w) c", b1=rows, b2=columns
    )
    if scale != 1:
        image_arr = np.kron(image_arr, np.ones((scale, scale, 1))).astype(np.uint8)
    image_grid = Image.fromarray(image_arr)
    return image_grid


def plthist(x, *args, **kwargs):
    x = tensor_to_numpy(x)
    plt.histogram(x, *args, **kwargs)
    plt.show()


def hist(x, info=True, mask=None, ignore_small=False, *args, **kwargs):
    x = tensor_to_numpy(x)
    if info is True:
        info = np.arange(len(x))

    if ignore_small:
        abs_x = np.abs(x)
        x_max = np.max(abs_x)
        small = abs_x < 0.01 * x_max
        if mask is None:
            mask = ~small
        else:
            new_mask = np.zeros_like(x, dtype=bool)
            new_mask[mask] = True
            mask = new_mask | ~small
            assert mask.dtype == bool, mask.dtype
            mask = mask | ~small

    if mask is not None:
        mask = tensor_to_numpy(mask)
        if info is not None:
            info = info[mask]
        x = x[mask]

    if info is not None:
        fig = px.histogram(x, marginal="rug", hover_name=info, *args, **kwargs)
    else:
        fig = px.histogram(x, *args, **kwargs)
    fig.update_layout(showlegend=False)

    return fig



from .hook_utils import *
from .rendering import (
    render_array,
    render_array_w_sign,
    get_image_grid,
    heatmap,
    qq_plot,
    str_arr_add,
    tensor_to_numpy,
    plthist,
    hist,
)
from .general_utils import *
from .seriation_utils import (
    get_local_distance_minimizing_permutation,
    get_seriation_permutations,
    seriate,
)
from .s3_utils import *



from tsp_solver.greedy_numpy import solve_tsp
import torch
import numpy as np
from functools import cache


# @cache
def get_local_distance_minimizing_permutation(x):
    """Get a permutation that minimizes the travelling distance between the rows of x"""
    assert len(x.shape) == 2, "x should be interpretable as a list of vecs"
    n_vectors = x.shape[0]
    distances = (x[None, :, :] - x[:, None, :]).norm(dim=-1)
    distances = torch.cat((distances, torch.zeros(1, n_vectors)), axis=0)
    distances = torch.cat((distances, torch.zeros(n_vectors + 1, 1)), axis=1)
    tour = solve_tsp(distances.numpy())
    permutation = torch.tensor(
        tour[tour.index(n_vectors) + 1 :] + tour[: tour.index(n_vectors)]
    )
    return permutation


# @cache
def get_seriation_permutations(x):
    perm_1, perm_2 = get_local_distance_minimizing_permutation(
        x
    ), get_local_distance_minimizing_permutation(x.T)
    return perm_1, perm_2


# @cache
def seriate(x):
    perm_1, perm_2 = get_local_distance_minimizing_permutation(
        x
    ), get_local_distance_minimizing_permutation(x.T)
    return x[perm_1][:, perm_2], perm_1, perm_2



from threading import Thread
from queue import Queue
import openai
import os

def threaded_map(fn, input, n_threads=20):
    input_queue = Queue()
    for idx, item in enumerate(input):
        input_queue.put((idx, item))
    results_queue = Queue()

    def worker():
        while not input_queue.empty():
            idx, item = input_queue.get()
            result = fn(item)
            results_queue.put((idx, result))
            input_queue.task_done()
            # time.sleep(1)
    
    for _ in range(n_threads):
        thread = Thread(target=worker)
        thread.daemon = True
        thread.start()
    
    input_queue.join()
    
    results = []
    while not results_queue.empty():        
        results.append(results_queue.get())

    return [result for idx, result in sorted(results, key=lambda x: x[0])]

openai.api_key = os.getenv("OPENAI_API_KEY")

def threaded_query(system_prompt, messages, max_tokens=100, model="gpt-3.5-turbo", n_threads=10):
    def query(message):
        response = openai.ChatCompletion.create(
            model=model,
            messages = [{'role': 'system', 'content': system_prompt}, {'role': 'user', 'content': message}],
            temperature=1.0,
            max_tokens=max_tokens,
        )
        return response['choices'][0]['message']['content']
    n_threads = min(n_threads, len(messages))
    return threaded_map(query, messages, n_threads=n_threads)


import logging
import boto3
from botocore.exceptions import ClientError
import os


def upload_file(
    file_name,
    bucket,
    object_name=None,
    public=False,
    access_key=None,
    secret_access_key=None,
    content_type="text/html"
):
    """Upload a file to an S3 bucket

    :param file_name: File to upload
    :param bucket: Bucket to upload to
    :param object_name: S3 object name. If not specified then file_name is used
    :return: True if file was uploaded, else False
    """

    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = os.path.basename(file_name)

    # Upload the file
    if access_key is None and secret_access_key is None:
        s3_client = boto3.client("s3")
    else:
        assert access_key is not None and secret_access_key is not None
        s3_client = boto3.client(
            "s3", aws_access_key_id=access_key, aws_secret_access_key=secret_access_key
        )

    try:
        if public is True:
            s3_client.upload_file(
                file_name,
                bucket,
                object_name,
                ExtraArgs={
                    "ACL": "public-read",
                    "ContentDisposition": "inline",
                    "ContentType": content_type,
                },
            )
        else:
            s3_client.upload_file(file_name, bucket, object_name)
    except ClientError as e:
        logging.error(e)
        return False
    url = f"https://{bucket}.s3.amazonaws.com/{object_name}"
    return url


html_head = """<head>
	<link rel="stylesheet"
          href="https://fonts.googleapis.com/css?family=Proxima_Nova">
	<style>
		body {
			font-family: 'Proxima_Nova', sans-serif;
		}
		p {
			font-size: 1.3rem;
		}
		h2 {
			font-size: 1.6rem
		}
	</style>
</head>

"""


def upload_figs(
    figs,
    fname,
    bucket="plotly-figs",
    fig_info=[],
    public=True,
    access_key=None,
    secret_access_key=None,
):
    """Uploads a plotly figure to an S3 bucket
    Returns S3 url if upload is successful, otherwise returns False
    """
    if not isinstance(figs, list):
        figs = [figs]
    if not isinstance(fig_info, list):
        fig_info = [fig_info]
    if len(fig_info) == 0:
        fig_info = ["" for i in range(len(figs))]
    assert len(figs) == len(fig_info)

    # assert fname.split('.')[-1] == 'html', 'fname must end in .html'

    res_html = html_head
    for fig, info in zip(figs, fig_info):
        res_html += "<body>" + info + "</body>"
        res_html += fig.to_html(full_html=False, include_plotlyjs="cdn")
    with open(f"/tmp/{fname}", "w") as f:
        f.write(res_html)
    url = upload_file(
        "/tmp/" + fname,
        bucket,
        object_name=fname,
        public=public,
        access_key=access_key,
        secret_access_key=secret_access_key,
    )
    os.remove("/tmp/" + fname)
    return url


def upload_pysvelte(
    fig,
    fname,
    bucket="plotly-figs",
    public=True,
    access_key=None,
    secret_access_key=None,
):
    """Uploads a pysvelte component to an S3 bucket
    Returns S3 url if upload is successful, otherwise returns False
    """
    # if not isinstance(figs, list):
    #     figs = [figs]
    # if not isinstance(fig_info, list):
    #     fig_info = [fig_info]
    # if len(fig_info) == 0:
    #     fig_info = ["" for i in range(len(figs))]
    # assert len(figs) == len(fig_info)

    # assert fname.split('.')[-1] == 'html', 'fname must end in .html'

    # res_html = html_head
    # res_html = ""
    # for fig, info in zip(figs, fig_info):
    #     # res_html += "<body>" + info + "</body>"
    #     res_html += fig.html_str()
    # with open(f"/tmp/{fname}", "w") as f:
    #     f.write(res_html)
    if not fname.endswith(".html"):
        fname += ".html"
    fig.publish(f"/tmp/{fname}")
    url = upload_file(
        "/tmp/" + fname,
        bucket,
        object_name=fname,
        public=public,
        access_key=access_key,
        secret_access_key=secret_access_key,
    )
    os.remove("/tmp/" + fname)
    return url



import torch
from typing import *
from types import MethodType


def _prepare_module(module):
    if not hasattr(module, "cache"):
        module.cache = {}

    if not hasattr(module, "hooks"):
        module.hooks = {}

    for child in module.children():
        _prepare_module(child)


def prepare_module_for_hook(module, hook_fn):
    _prepare_module(module)

    if hook_fn.__name__ not in module.hooks:
        return True
    else:
        print(
            "hook_fn already registered!:",
            module.__class__.__name__,
            hook_fn.__name__,
        )
        return False


def register_hook(
    module,
    hook_fn: Callable,
):
    if prepare_module_for_hook(module, hook_fn) is False:
        return

    removable_handle = module.register_forward_hook(
        lambda module, input, output: hook_fn(module, input, output)
    )
    module.hooks[hook_fn.__name__] = removable_handle

    return removable_handle


def register_backward_hook(module, hook_fn: Callable):
    if prepare_module_for_hook(module, hook_fn) is False:
        return

    removable_handle = module.register_full_backward_hook(
        lambda module, grad_input, grad_output: hook_fn(module, grad_input, grad_output)
    )
    module.hooks[hook_fn.__name__] = removable_handle
    return removable_handle


def recursively_delete(nest):
    if isinstance(nest, tuple) or isinstance(nest, list):
        for it in nest:
            recursively_delete(it)
    elif isinstance(nest, dict):
        for k, v in nest.items():
            recursively_delete(v)
    elif isinstance(nest, torch.Tensor):
        del nest
    else:
        del nest


def clear_cache(module):
    if hasattr(module, "cache"):
        recursively_delete(module.cache)
        module.cache = {}
    for child in module.children():
        clear_cache(child)


def remove_hooks(module, quiet=False, wipe_cache=True):
    # Recursively remove hooks from a module and its children
    if hasattr(module, "hooks"):
        if not isinstance(module.hooks, dict):
            module.hooks = {}
        for hook_name, removable_handle in module.hooks.items():
            if quiet is False:
                print("Removing hook: ", module.__class__.__name__, hook_name)
            removable_handle.remove()
        module.hooks = {}
    if wipe_cache is True:
        clear_cache(module)
    for child in module.children():
        remove_hooks(child, quiet=quiet, wipe_cache=wipe_cache)


def to_cpu(nest):
    if isinstance(nest, tuple):
        return tuple([to_cpu(it) for it in nest])
    if isinstance(nest, list):
        return [to_cpu(it) for it in nest]
    if isinstance(nest, torch.Tensor):
        # return nest.clone().detach().requires_grad_(True).cpu()
        return nest.cpu()
    if isinstance(nest, dict):
        return {k: to_cpu(v) for k, v in nest.items()}
    else:
        assert False, f"Unknown nest type: {type(nest)}"


def caching_hook(module, input, output):
    assert isinstance(input, tuple) and len(input) == 1
    input = input[0]
    module.cache["input"] = to_cpu(input)
    module.cache["output"] = to_cpu(output)


def include_keys(d: dict, keys: list):
    for k in keys:
        if not k in d:
            d[k] = []


def append_cache_hook(module, input, output):
    include_keys(module.cache, ["input", "output"])

    module.cache["input"].append(to_cpu(input))
    module.cache["output"].append(to_cpu(output))


def append_input_hook(module, input, output):
    include_keys(module.cache, ["input"])

    module.cache["input"].append(to_cpu(input))


def append_output_hook(module, input, output):
    include_keys(module.cache, ["output"])

    module.cache["output"].append(to_cpu(output))
